# -*- coding: utf-8 -*-
"""
Download and build FreeTDS library and build pymssql.
"""

import argparse
from distutils.util import get_platform
import multiprocessing
import os
from pathlib import Path
import platform
import shutil
from subprocess import check_call, check_output, STDOUT
import sys
import tarfile


def run(cmd, cwd=None, env=None, shell=True):

    check_call(str(cmd), cwd=str(cwd) if cwd else cwd, env=env, shell=shell, stderr=STDOUT)


def download(args):

    args.ws_dir.mkdir(parents=True, exist_ok=True)
    if args.freetds_version.lower() == 'latest':
        args.freetds_version ='patched'
    arcname = f"freetds-{args.freetds_version}.tar.gz"
    url = f"{args.freetds_url.strip('/')}/{arcname}"
    freetds_dest = args.ws_dir / arcname
    if freetds_dest.exists() and not args.force_download:
        print(f"{freetds_dest} already exists")
    else:
        print(f"downloading {url} to {freetds_dest}")
        run(f"curl -sSf --retry 10 --max-time 10  {url} -o {freetds_dest}")

    if platform.system() == 'Windows':
        url = f"{args.iconv_url.strip('/')}/v{args.iconv_version}"
        iconv_dest = args.ws_dir / "win-iconv.zip"
        if iconv_dest.exists() and not args.force_download:
            print(f"{iconv_dest} already exists")
        else:
            run(f"curl -sSf {url} -o {iconv_dest}")
    else:
        iconv_dest = None

    print("downloading DONE")
    return freetds_dest, iconv_dest


def build(args, freetds_archive):

    tar = tarfile.open(freetds_archive)
    topdir = tar.next().name
    srcdir = args.ws_dir / f"{topdir}"
    if srcdir.exists():
        shutil.rmtree(srcdir)
    tar.extractall(args.ws_dir)
    tar.close()
    if args.prefix is None:
        args.prefix = args.ws_dir / f"{topdir}-bin"

    blddir = args.ws_dir / f"{topdir}-build"
    if blddir.exists():
        shutil.rmtree(blddir)
    blddir.mkdir(parents=True)

    configure = srcdir / "configure"
    configure = [ f'"{configure}"',
                  f"--prefix={args.prefix}",
                  "--enable-msdblib",
                  f"--with-tdsver={args.with_tdsver}",
                  #"--disable-apps",
                  "--disable-server",
                  "--disable-pool",
                  "--disable-odbc",
                  f"--with-openssl={args.with_openssl}",
                  "--with-gnutls=no",
                  ]
    if args.enable_krb5:
        configure.append("--enable-krb5")
    if args.static_freetds:
        configure.extend(["--enable-static", "--disable-shared"])
    else:
        configure.extend(["--enable-shared", "--disable-static"])
    cmd = ' '.join(configure)
    env = os.environ.copy()
    env.update(CFLAGS="-fPIC")
    run(cmd, cwd=blddir, env=env)
    make = f"make -j {multiprocessing.cpu_count()}"
    run(make, cwd=blddir, env=env)
    make = "make install"
    run(make, cwd=blddir, env=env)


def find_vcvarsall_env():

    from distutils import _msvccompiler as _msvcc

    plat_name = get_platform()
    plat_spec = _msvcc.PLAT_TO_VCVARS[plat_name]
    vcvarsall, _ = _msvcc._find_vcvarsall(plat_spec)
    cmd = f'(call "{vcvarsall}" {plat_spec}>nul)&&"{sys.executable}" -c "import os;print(repr(os.environ))"'
    env = check_output(cmd, shell=True)
    env = eval(env.decode('ascii').strip('environ'))
    return env


def build_windows(args, freetds_archive, iconv_archive):

    from zipfile import ZipFile
    wiconv = args.ws_dir / "win-iconv"
    if wiconv.exists():
        shutil.rmtree(wiconv)
    wiconv.mkdir(parents=True)
    print(f"extracting {iconv_archive.name} -> {wiconv}")
    with ZipFile(iconv_archive) as zipf:
        for m in zipf.namelist():
            fn = m.split('/', 1)[1]
            if fn:
                (wiconv / fn).write_bytes(zipf.read(m))

    env = find_vcvarsall_env()

    cmd = f'"{args.cmake}" -G "NMake Makefiles" ' \
            '-DCMAKE_BUILD_TYPE=Release -DBUILD_STATIC=on -DBUILD_SHARED=off ' \
            '-DBUILD_EXECUTABLE=off -DBUILD_TEST=off ' \
            '.'
    print(f"running cmake in {wiconv}")
    run(cmd, cwd=wiconv, env=env)
    print(f"running nmake in {wiconv}")
    run("nmake", cwd=wiconv, env=env)

    tar = tarfile.open(freetds_archive)
    topdir = tar.next().name
    srcdir = args.ws_dir / f"{topdir}"
    if srcdir.exists():
        shutil.rmtree(srcdir)
    tar.extractall(args.ws_dir)
    tar.close()
    if args.prefix is None:
        args.prefix = args.ws_dir / f"{topdir}-bin"

    blddir = args.ws_dir / f"{topdir}-build"
    if blddir.exists():
        shutil.rmtree(blddir)
    blddir.mkdir(parents=True)

    os.makedirs(blddir / "lib", exist_ok=True)
    shutil.copy(wiconv / "iconv.h", blddir)
    shutil.copy(wiconv / "iconv.lib", blddir / "lib")

    krb5 = "ON" if args.enable_krb5 else "OFF"
    cmd = f'"{args.cmake}" -G "NMake Makefiles" -DCMAKE_BUILD_TYPE=Release " \
            "-DWITH_OPENSSL=ON -DENABLE_KRB5={krb5} -DCMAKE_INSTALL_PREFIX="{args.prefix}" "{srcdir}"'
    env["PATH"] += f";{args.msys}"
    run(cmd, cwd=blddir, env=env)

    run("nmake install", cwd=blddir, env=env)


def freetds_version(freetds_version):

    if freetds_version:
        return freetds_version

    freetds_version = "latest"

    pyproject_toml = Path(__file__).parent.parent / "pyproject.toml"
    if not pyproject_toml.exists():
        print("pyproject.toml not found, use latest FreeTDS")
        return freetds_version

    try:
        from tomllib import loads
    except ModuleNotFoundError:
        from tomli import loads
    except ModuleNotFoundError:
        print("Neither tomllib no tomli module found, try latest FreeTDS.")
        return freetds_version

    pyproject = loads(pyproject_toml.read_text())
    try:
        freetds_version = pyproject['tool']['freetds']['version_for_pypi_wheels']
    except:
        pass
    return freetds_version


def parse_args(argv):

    parser = argparse.ArgumentParser(description=(__doc__))
    a = parser.add_argument

    a('-f', '--force-download', action='store_true',
            help="force archive download")
    a('-u', '--freetds-url', default="http://ftp.freetds.org/pub/freetds/stable",
            help="URL to download FreeTDS archive")
    a('-v', '--freetds-version', default='', type=freetds_version,
            help="""Specific FreeTDS version to build or 'latest'.
                    If not set, try to read from pyproject.toml""")
    if platform.system() != 'Windows':
        a('-t', '--with-tdsver', choices=['5.0', '7.1', '7.2', '7.3', '7.4', 'auto'], default='auto',
                help="TDS protocol version")
        a('-o', '--with-openssl', choices=['yes', 'no'], default='yes',
                help="Build FreeTDS with or without OpenSSL support, default is 'yes'")
        a('-S', '--static-freetds', action='store_true',
                help="build FreeTDS staticly")
    a('-k', '--enable-krb5', dest="enable_krb5", action='store_true',
            help="enable krb5 support")

    base = Path('~/freetds').expanduser()
    a('-w', '--ws-dir', default=base, type=Path,
            help="workspace directory for building FreeTDS")
    a('-p', '--prefix', default=None, type=lambda x: Path(x) if x else None,
            help="prefix for installing FreeTDS, default is WS_DIR/prefix")
    a('-d', '--dist-dir', default=Path('./dist'), type=Path,
            help="where to put pymssql wheel, default is './dist'")
    a('-s', '--sdist', action='store_true',
            help="build sdist archive")

    if platform.system() == 'Windows':
        a('-U', '--iconv-url', default="https://codeload.github.com/win-iconv/win-iconv/zip",
                help="URL to download win-iconv.zip archive for build on Windows")
        a('-V', '--iconv-version', default='0.0.8',
                help="FreeTDS version to build")
        a('-m', '--msys', type=Path, default=Path("c:/tools/msys64/usr/bin"),
                help="Msys binaries installation directory")
        a('--cmake', type=Path, default=Path("C:/Program Files/CMake/bin/cmake.exe"),
                help="cmake executable for building FreeTDS")

    args = parser.parse_args(argv)
    return args


def main(argv):

    args = parse_args(argv)
    args.ws_dir = args.ws_dir.absolute()
    if args.prefix is not None:
        args.prefix = args.prefix.absolute()

    freetds_archive, iconv_archive = download(args)

    if platform.system() == 'Windows':
        os.environ["PATH"] += f";{args.msys}"
        build_windows(args, freetds_archive, iconv_archive)
    else:
        build(args, freetds_archive)

    args.dist_dir = args.dist_dir.absolute()
    env = os.environ.copy()
    env.update(PYMSSQL_FREETDS=f"{args.prefix}")
    run(f"{sys.executable} -m pip wheel . -w {args.dist_dir}", shell=True, env=env)
    if args.sdist:
        fmt = 'zip' if platform.system() == 'Windows' else 'gztar'
        run(f"{sys.executable} setup.py sdist --formats={fmt} -d {args.dist_dir}", shell=True, env=env)


if __name__ == '__main__':

    try:
        main(sys.argv[1:])
    except Exception as exc:
        print(f"Build failed: {exc}")
        sys.exit(1)




# -*- coding: utf-8 -*-
"""
Test charset usage in queries.
"""

import unittest

import pytest

from .helpers import pymssqlconn


@pytest.mark.mssql_server_required
class TestCharset(unittest.TestCase):

    def setUp(self):
        self.conn = pymssqlconn(charset='WINDOWS-1251')

    def test_charset(self):
        cursor = self.conn.cursor()

        try:
            cursor.execute(
                'select %s, %s',
                ('Здравствуй', 'Мир')  # Russian strings
            )
        except UnicodeDecodeError as e:
            self.fail("cursor.execute() raised %s unexpectedly: %s" % (e.__class__.__name__, e))

        a, b = cursor.fetchone()

        self.assertEqual(a, 'Здравствуй')
        self.assertEqual(b, 'Мир')



"""
This type stub file was generated by pyright.
"""

class Warning(Exception): ...
class Error(Exception): ...
class InterfaceError(Error): ...
class DatabaseError(Error): ...
class DataError(DatabaseError): ...
class OperationalError(DatabaseError): ...
class IntegrityError(DatabaseError): ...
class InternalError(DatabaseError): ...
class ProgrammingError(DatabaseError): ...
class NotSupportedError(DatabaseError): ...

class ColumnsWithoutNamesError(InterfaceError):
    def __init__(self, columns_without_names) -> None: ...



# -*- coding: utf-8 -*-
"""
Test unicode usage in queries.
"""

import unittest

import pytest

from .helpers import pymssqlconn

@pytest.mark.mssql_server_required
class TestUnicode(unittest.TestCase):

    def setUp(self):
        self.conn = pymssqlconn()

    def test_unicode(self):
        # This test is for http://code.google.com/p/pymssql/issues/detail?id=60
        # Thanks to tonal.promsoft for reporting the issue and submitting a
        # patch.

        cursor = self.conn.cursor()
        cursor.execute(
            'select %s, %s',
            ('Здравствуй',
             'Мир'))  # Russian strings

        a, b = cursor.fetchone()

        self.assertEqual(a, 'Здравствуй')
        self.assertEqual(b, 'Мир')



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# setup.py
#
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301  USA
#

import os
from os.path import exists, join, splitext
from pathlib import Path
import platform
import struct
import subprocess
import sys

from setuptools import setup, Extension

from distutils import log
from distutils.cmd import Command
from distutils.command.clean import clean as _clean

have_c_files = exists('pymssql/_mssql.c') and exists('pymssql/_pymssql.c')
if have_c_files:
    from distutils.command.build_ext import build_ext as _build_ext
else:
    # Force `setup_requires` stuff like Cython to be installed before proceeding
    from setuptools.dist import Distribution
    Distribution(dict(setup_requires='Cython>=3.0'))
    from Cython.Distutils import build_ext as _build_ext

def check_env(env_name, default):
    val = os.getenv(env_name, default)
    if val.upper() in ('1', 'YES', 'TRUE'):
       return True
    elif val.upper() in ('0', 'NO', 'FALSE'):
       return False
    else:
       raise Exception(f"Unsupported environment value {env_name}={val}")

LINK_FREETDS_STATICALLY = check_env('LINK_FREETDS_STATICALLY', 'YES')
LINK_OPENSSL = check_env('LINK_OPENSSL', 'YES')
LINK_KRB5 = check_env('LINK_KRB5', 'YES')

# 32 bit or 64 bit system?
BITNESS = struct.calcsize("P") * 8
WINDOWS = platform.system() == 'Windows'
MACOS = platform.system() == 'Darwin'

include_dirs = []
library_dirs = []
libraries = ['sybdb']

prefix = "/usr/local"
if os.getenv('PYMSSQL_FREETDS'):
    prefix = os.path.abspath(os.getenv('PYMSSQL_FREETDS').strip())
elif exists("/usr/local/include/sqlfront.h"):
    prefix = "/usr/local"
elif exists("/usr/local/opt/freetds/include/sqlfront.h"): # brew macOS on Intel
    prefix = "/usr/local/opt/freetds"
elif exists("/opt/homebrew/opt/freetds/include/sqlfront.h"): # brew macOS on Apple Silicon/ARM
    prefix = "/opt/homebrew/opt/freetds"
elif exists("/opt/local/include/sqlfront.h"): # MacPorts
    prefix = "/opt/local"
elif exists("/sw/include/sqlfront.h"): # Fink
    prefix = "/sw"
print(f"prefix='{prefix}'")

if os.getenv('PYMSSQL_FREETDS_INCLUDEDIR'):
    include_dirs = [ os.getenv('PYMSSQL_FREETDS_INCLUDEDIR') ]
else:
    include_dirs = [ join(prefix, "include") ]

if os.getenv('PYMSSQL_FREETDS_LIBDIR'):
    library_dirs = [ os.getenv('PYMSSQL_FREETDS_LIBDIR') ]
else:
    if BITNESS == 64 and exists(join(prefix, "lib64")):
        library_dirs = [ join(prefix, "lib64") ]
    else:
        library_dirs = [ join(prefix, "lib") ]

if MACOS:
    pref = subprocess.getoutput('brew --prefix openssl')
    print(f"PREFIX={pref}")
    include_dirs.append(f"{pref}/include")
    library_dirs.append(f"{pref}/lib")

print("setup.py: platform.system() =>", platform.system())
print("setup.py: platform.architecture() =>", platform.architecture())
if not WINDOWS:
    print("setup.py: platform.libc_ver() =>", platform.libc_ver())
print("setup.py: include_dirs =>", include_dirs)
print("setup.py: library_dirs =>", library_dirs)

if not WINDOWS and platform.libc_ver()[0] == 'glibc':
    # check for clock_gettime, link with librt for glibc<2.17
    from dev import ccompiler
    ccompiler.check_clock_gettime(libraries)


class build_ext(_build_ext):
    """
    Subclass the Cython build_ext command so it:
    * Can handle different C compilers on Windows
    * Links in the libraries we collected
    """

    def build_extensions(self):
        global library_dirs, include_dirs, libraries

        extra_cc_args = []
        if WINDOWS:
            # Detect the compiler so we can specify the correct command line switches
            # and libraries
            from distutils.cygwinccompiler import Mingw32CCompiler
            if isinstance(self.compiler, Mingw32CCompiler):
                # Compiler is Mingw32
                extra_cc_args = [
                    '-Wl,-allow-multiple-definition',
                    '-Wl,-subsystem,windows-mthreads',
                    '-mwindows',
                    '-Wl,--strip-all'
                ]
                libraries = [
                    'libiconv', 'iconv',
                    'sybdb',
                    'ws2_32', 'wsock32', 'kernel32',
                ]
            else:
                # Assume compiler is Visual Studio
                if LINK_FREETDS_STATICALLY:
                    libraries = [
                        'replacements',
                        'db-lib', 'tds', 'tdsutils',
                        'ws2_32', 'wsock32', 'kernel32', 'shell32',
                    ]
                    if LINK_OPENSSL:
                        libraries.extend([
                            'libssl_static', 'libcrypto_static',
                            'crypt32', 'advapi32', 'gdi32', 'user32',
                        ])
                else:
                    libraries = [
                        'sybdb',
                        'ws2_32', 'wsock32', 'kernel32', 'shell32',
                    ]
                    if LINK_OPENSSL:
                        libraries.extend(['libssl', 'libcrypto'])

            for e in self.extensions:
                e.extra_compile_args.extend(extra_cc_args)
                e.libraries.extend(libraries)
                if LINK_OPENSSL:
                    if BITNESS == 32:
                        e.library_dirs.append("c:/Program Files (x86)/OpenSSL-Win32/lib")
                    else:
                        e.library_dirs.append("c:/Program Files/OpenSSL/lib")

        else:
            if LINK_KRB5:
                libraries.extend([ 'gssapi_krb5', 'krb5'] )
            if LINK_OPENSSL and LINK_FREETDS_STATICALLY:
                libraries.extend([ 'ssl', 'crypto' ])
            if MACOS:
                extra_cc_args = [
                    '-Wno-unreachable-code-fallthrough',
                ]
            for e in self.extensions:
                e.libraries.extend(libraries)
                e.extra_compile_args.extend(extra_cc_args)

        _build_ext.build_extensions(self)


class clean(_clean):
    """
    Subclass clean so it removes all the Cython generated files.
    """

    def run(self):
        _clean.run(self)
        for ext in self.distribution.ext_modules:
            cy_sources = [splitext(s)[0] for s in ext.sources]
            for cy_source in cy_sources:
                # .so/.pyd files are created in place when using 'develop'
                for ext in ('.c', '.so', '.pyd'):
                    generated = cy_source + ext
                    if exists(generated):
                        log.info('removing %s', generated)
                        os.remove(generated)


class release(Command):
    """
    Setuptools command to run all the required commands to perform
    a release. This acts differently depending on the platform it
    is being run on.
    """

    description = "Run all the commands required for a release."

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if WINDOWS:
            self.release_windows()
        else:
            self.release_unix()

    def release_windows(self):
        # generate windows source distributions
        sdist = self.distribution.get_command_obj('sdist')
        sdist.formats = 'zip'
        sdist.ensure_finalized()
        sdist.run()

        # generate a windows egg
        self.run_command('bdist_egg')

        # generate windows installers
        bdist = self.reinitialize_command('bdist')
        bdist.formats = 'zip,wininst'
        bdist.ensure_finalized()
        bdist.run()

    def release_unix(self):
        # generate linux source distributions
        sdist = self.distribution.get_command_obj('sdist')
        sdist.formats = 'gztar,bztar'
        sdist.ensure_finalized()
        sdist.run()

def ext_modules():
    if have_c_files:
        source_extension = 'c'
    else:
        source_extension = 'pyx'

    ext_modules = [
        Extension('pymssql._mssql', [join('src', 'pymssql', '_mssql.%s' % source_extension)],
            extra_compile_args = [ '-DMSDBLIB' ],
            include_dirs = include_dirs,
            library_dirs = library_dirs,
        ),
        Extension('pymssql._pymssql', [join('src', 'pymssql', '_pymssql.%s' % source_extension)],
            extra_compile_args = [ '-DMSDBLIB' ],
            include_dirs = include_dirs,
            library_dirs = library_dirs,
        ),
    ]
    for e in ext_modules:
        e.cython_directives = {'language_level': sys.version_info[0]}
    return ext_modules


def mk_long_description(numrev=1):

    readme = (Path('__file__').parent / 'README.rst').read_text()
    chlog = Path('__file__').parent / 'ChangeLog.rst'
    lines = []
    with chlog.open('r') as f:
        count = 0
        l = f.readline()
        while l:
            if l.startswith('Version 2'):
                count += 1
            if count > numrev:
                break
            lines.append(l)
            l = f.readline()
    return readme + "\n\n" + ''.join(lines).strip()


setup(
    name  = 'pymssql',
    use_scm_version = {
        "write_to": "src/pymssql/version.h",
        "write_to_template": 'const char* PYMSSQL_VERSION = "{version}";',
        "local_scheme": "no-local-version",
    },
    description = 'DB-API interface to Microsoft SQL Server for Python. (new Cython-based version)',
    long_description = mk_long_description(2),
    long_description_content_type = 'text/x-rst',
    author = 'Damien Churchill',
    author_email = 'damoxc@gmail.com',
    maintainer = 'Mikhail Terekhov',
    maintainer_email = 'termim@gmail.com',
    license = 'LGPL',
    platforms = 'any',
    keywords = ['mssql', 'SQL Server', 'database', 'DB-API'],
    project_urls={
        "Documentation": "http://pymssql.readthedocs.io",
        "Source": "https://github.com/pymssql/pymssql",
        "Changelog": "https://github.com/pymssql/pymssql/blob/master/ChangeLog.rst",
    },
    cmdclass = {
        'build_ext': build_ext,
        'clean': clean,
        'release': release,
    },
    classifiers=[
      "Development Status :: 5 - Production/Stable",
      "Intended Audience :: Developers",
      "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
      "Programming Language :: Python",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "Programming Language :: Python :: 3.11",
      "Programming Language :: Python :: 3.12",
      "Programming Language :: Python :: Implementation :: CPython",
      "Topic :: Database",
      "Topic :: Database :: Database Engines/Servers",
      "Topic :: Software Development :: Libraries :: Python Modules",
      "Operating System :: Microsoft :: Windows",
      "Operating System :: POSIX",
      "Operating System :: Unix",
    ],
    zip_safe = False,
    setup_requires=['setuptools_scm[toml]>=5.0,<9.0', 'Cython>=3.0.7'],
    tests_require=['psutil<5.9.9', 'pytest', 'pytest-timeout', 'pytest-subtests'],
    ext_modules = ext_modules(),
    packages = [ 'pymssql'],
    package_dir = {'': 'src'},
)



# -*- coding: utf-8 -*-

import contextlib
from distutils import ccompiler
from distutils.errors import *
import os
import sys
import types


def _has_function(self, funcname, includes=None, include_dirs=None,
                 libraries=None, library_dirs=None):
    """Return a boolean indicating whether funcname is supported on
    the current platform.  The optional arguments can be used to
    augment the compilation environment.
    """
    # this can't be included at module scope because it tries to
    # import math which might not be available at that point - maybe
    # the necessary logic should just be inlined?
    import tempfile
    if includes is None:
        includes = []
    if include_dirs is None:
        include_dirs = []
    if libraries is None:
        libraries = []
    if library_dirs is None:
        library_dirs = []
    fd, fname = tempfile.mkstemp(".c", funcname, text=True)
    f = os.fdopen(fd, "w")
    try:
        for incl in includes:
            f.write("""#include "%s"\n""" % incl)
        f.write("""\
int main (int argc, char **argv) {
%s;
return 0;
}
""" % funcname)
    finally:
        f.close()
    try:
        objects = self.compile([fname], include_dirs=include_dirs)
    except CompileError:
        return False
    finally:
        os.remove(fname)

    try:
        self.link_executable(objects, "a.out",
                             libraries=libraries,
                             library_dirs=library_dirs)
    except (LinkError, TypeError):
        return False
    else:
        os.remove("a.out")
    finally:
        for fn in objects:
            os.remove(fn)

    return True


@contextlib.contextmanager
def stdchannel_redirected(stdchannel, dest_filename):
    """
    A context manager to temporarily redirect stdout or stderr

    e.g.:

    with stdchannel_redirected(sys.stderr, os.devnull):
        ...
    """

    try:
        oldstdchannel = os.dup(stdchannel.fileno())
        dest_file = open(dest_filename, 'w')
        os.dup2(dest_file.fileno(), stdchannel.fileno())

        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, stdchannel.fileno())
        if dest_file is not None:
            dest_file.close()


def has_function(*args, **kw):

        with stdchannel_redirected(sys.stderr, os.devnull):
            return _has_function(*args, **kw)


def new_compiler():

    compiler = ccompiler.new_compiler()
    compiler.has_function = types.MethodType(has_function, compiler)
    return compiler


def find_compiler():

    from subprocess import run, STDOUT, PIPE

    for cc in ['cc', 'gcc']:
        x = run("which %s" % cc, stdout=PIPE, stderr=STDOUT, shell=True)
        if x.returncode == 0:
            return True
    return False


def check_clock_gettime(libraries):
    """
        check for clock_gettime, link with librt for glibc<2.17
    """
    compiler = new_compiler()
    try:
        if not compiler.has_function('clock_gettime(0,NULL)', includes=['time.h']):
            if compiler.has_function('clock_gettime(0,NULL)', includes=['time.h'], libraries=['rt']):
                libraries.append('rt')
            else:
                if find_compiler():
                    print("setup.py: could not locate 'clock_gettime' function required by FreeTDS.", file=sys.stderr)
                else:
                    print("setup.py: ERROR: Could not find C compiler", file=sys.stderr)
                sys.exit(1)
    except Exception as exc:
        print(f"setup.py: ERROR: {exc}", file=sys.stderr)
        sys.exit(1)
